-- Test routines to see whether code returned by 
-- allocate_code() and call_back() crash.  They should never crash.
--
-- code called into memory returned by allocate() will crash if
-- DEP is enabled.
--
-- Now, if calling code using memory returned by allocate() doesn't
-- crash the process we consider the test inconclusive.

-- Flow control
-- 
-- The idea is we try to do things that might cause the program
-- to crash and then the crash handler puts test_fail() for each
-- thing we tried.  In the case of running code from memory returned
-- by allocate() we know the other things did not crash so we
-- call test_pass there.

include xalloc.e as dep
include joy.e as joy
include std/filesys.e as fs
include machine.e as mach 
include dll.e as dll
 

ifdef UNITTEST then
	include std/unittest.e
	override procedure puts( integer f, object s )
	end procedure
	override procedure printf( integer fd, sequence fmt, sequence s )
	end procedure
end ifdef
ifdef SILENCE then
	override procedure puts( integer f, object s )
	end procedure
	override procedure printf( integer fd, sequence fmt, sequence s )
	end procedure
	
end ifdef
function isspace( integer x )
	return find( x, "\t \n" )
end function

integer log
log = open( "wordwrap.log", "w" )

-- word_wrap function:  Please do not pass strings with TAB characters
function word_wrap( joy:string_of_integers s, integer line_length )
	sequence lines
	sequence haystack	
	integer whitelocation
	
	lines = {}
	while length(s) do
		haystack = s[1..joy:min({length(s),line_length})]
		puts( log, "haystack = "&haystack&10 )
		if length(s) <= line_length then
			lines = append( lines, s )
			s = ""
		else -- length > line_length
			-- if no spaces Grab line_length of it
			-- if '\n' grab the first one, or the last ' ' 
			-- whichever comes first.
			whitelocation = joy:min( {length(s) + 1} & joy:remove_0s( { find( '\n', haystack ), joy:rfind( ' ', haystack ) } ) )

			puts( log, "s[1..whitelocation-1]=" & s[1..whitelocation-1] & 10 )
			lines = append( lines, s[1..whitelocation-1] )
			flush(log)
			

			if whitelocation>=length(s) then
			    s = ""
			else
			    s = s[whitelocation+1..length(s)]
			end if
			
		end if
	end while
	return joy:join( "\n", lines )
end function

object line


procedure PETC()
	ifdef !UNITTEST then
		puts(1,"Press Enter to Continue")
		line = gets(0)
		puts(1,10)
	end ifdef
end procedure

function fPETC()
	ifdef !UNITTEST then
		puts(1,"Press Enter to Continue")
		line = gets(0)
		puts(1,10)
		return line
	elsedef
		return "test\n"
	end ifdef
end function

integer calling_memfunction
calling_memfunction = 0
function print_failure(object x)
    if calling_memfunction then
        puts( 1, word_wrap( "  We have found that the xalloc.e routines do not really make memory executable.\n", 80 ) )
	puts( 1, word_wrap( "This conclusively tells us that there is a bug in the allocate_code() function.  Please let me know about this:  Email me at shawn.pringle@gmail.com", 80 ) )
	puts( 1, "Test result for allocate_code: FAILURE\n" )
	PETC()
	ifdef UNITTEST then
		test_fail( "Is memory allocated by allocate_code() executable?" )
		test_report()
	end ifdef
        abort(0)
    end if
    return 0
end function

integer calling_dmemfunction
calling_dmemfunction = 0
function dep_is_enabled(object x)
    if calling_dmemfunction then
	if fs:delete_file( "ex.err" ) then
		puts(1, "The file ex.err has been fs:deleted.\n" )
	end if
        puts(1, word_wrap( "  The code allocated with allocate(), however raised an exception and it SHOULD.  Therefore D.E.P. is enabled for the interpreter you used for this test on your system.  This means, your system is configured correctly to test allocate_code() and that allocate_code() and call_back() work properly.\n", 80 ) )
	puts(1, "Test result for allocate_code: SUCCESS!\n" )
	ifdef UNITTEST then
		test_pass( "Is memory allocated by allocate_code() executable?" )
		test_pass( "Can we call functions returned by call_back?" )
		test_report()
	end ifdef
        PETC()
	abort(0)
    end if
    return 0
end function

integer calling_call_back = 0
function when_callback_call_fails( object x )
    if calling_call_back then
	if fs:delete_file( "ex.err" ) then
	end if
	ifdef UNITTEST then
		test_pass( "Is memory allocated by allocate_code() executable?" )
		test_fail( "Can we call functions returned by call_back?" )
		test_report()
	end ifdef
	puts( 1, "Executing from an address returned by callback caused an exception.\n" )
	PETC()
	abort(0)
    end if
    return 0
end function

atom code_space, data_space
sequence multiply_code
atom rexec, rdata
atom x,y
object void


-- machine code taken from callmach.ex
multiply_code = {
   -- int argument is at stack offset +4, double is at +8 
   #DB, #44, #24, #04,        -- fild  dword ptr +4[esp]
   #DC, #4C, #24, #08,        -- fmul  qword ptr +8[esp]
   #C2, #0C - 4 * (platform()=LINUX), #00  -- ret C -- pop 12 (or 8) bytes 
					   -- off the stack
    }

crash_routine(routine_id("print_failure"))
crash_routine(routine_id("dep_is_enabled"))
crash_routine(routine_id("when_callback_call_fails"))
code_space = dep:allocate_code(multiply_code)
data_space = allocate(length(multiply_code))
if data_space = 0 then
    puts(  1, word_wrap( "Could not allocate memory!", 80 ) )
    PETC()
    abort(0)
end if
poke( data_space, multiply_code )
rexec = define_c_func("", code_space, {C_INT, C_DOUBLE}, C_DOUBLE)
rdata = define_c_func("", data_space, {C_INT, C_DOUBLE}, C_DOUBLE ) 

x = 7
y = 8.5

-- Let's not confuse the user:  If they are running DOS32, Linux or FreeBSD they might wonder why
-- the test is 'inconclusive' without reading the help text.  Sometimes one can run in DOS mode
-- from WINDOWS without realizing it!
if platform() > LINUX then
	puts(1, word_wrap( "  This platform is new for this program.  Time for you to download an upgrade.\n", 80 ) )
	abort(0)
elsif platform() != WIN32 then
	puts(1, word_wrap( "  Currently as of 2008, the only OSes that support DEP and that EUPHORIA runs on are some versions of Windows and Intel-Mac OS X."  &
	"  You are not using Windows so this test would simply be inconclusive."&
	"  Try again, with another OS."&
	"  To run the test anyway type test and press ENTER or simply Press ENTER to quit.", 80) &
	"\n"
	)
	line = fPETC()
	if compare( line, "test\n" ) != 0 then
		abort(0)
	end if
end if

puts(  1,  word_wrap( "  We will now call a machine code function using memory marked as executable.\n", 80 )  )
PETC()
calling_memfunction = 1
void = c_func(rexec, {x, y})
calling_memfunction = 0
printf(1, word_wrap( "  The code allocated with allocate_code() was called without an exception being raised.  The function allocate_code() works as it should.\n", 80 ), {} )
puts(1, "\n\n" )

function ifthenelse( integer condition, sequence s1, object x2 )
	if condition then return s1 else return x2 end if
end function

function five() 
	return 5 
end function 
 
atom five_cb = call_back(routine_id("five"))  
integer cb = define_c_func("", {'+', five_cb}, {}, C_INT) 

printf(1, word_wrap( "  We will now try to call an address returned by call_back().\n", 80 ), {} )
calling_call_back = 1
void = c_func(cb, {}) 
calling_call_back = 0
PETC()
puts(1, word_wrap( "  The call_back() address was executed without an exception being raised.", 80 ) & 10 & 10 )

puts(1, word_wrap( "  We will now call a machine code function using normal memory.  If D.E.P is enabled this should cause an exception.\n", 80 ) )
PETC()
calling_dmemfunction = 1
void = c_func(rdata, {x,y})
calling_dmemfunction = 0
printf(1, word_wrap( "\n  The code allocated with allocate() was also called " &
    "without an exception being "&
    "raised.  This means your DEP is disabled for, at least, EUPHORIA." &
    "  " & 
   
    "You do not require DEP to use this library or this test.  "&
    "However, the purpose of these routines are so that you can run code on " & 
    "the stack when DEP is enabled.  You can always do so when DEP is disabled.",
    80
), {} )
    
    
	     
    -- "You probably can get into the pre-OS configuration "&
    -- "screen by pressing the fs:delete key at boot up.\n"&
    -- 
    -- ifthenelse( platform() = WIN32,
    -- "  In Windows XP, you can find the configuration for "&
    -- "hardware DEP by first, opening Control_Panel, next from there opening "&
    -- "System, after that clicking the Advanced_Tab, and next clicking the Performance button and "&
    -- "finally clicking the Data execution Prevention tab. " &
    -- "There you can choose the option for enabling DEP for all processes.", "")&
    -- 
    -- ifthenelse( platform() = LINUX, 
    -- "  In Linux, you will need to download some kernel patches to allow, "&
    -- "this to work.", "" ) & 
    -- 
    -- ifthenelse( platform() = DOS32, 
    -- "  In DOS, there is no way to enable hardware DEP.", "" )&"\n", 80)
    -- , {} )
close(log)




puts(1, "\nTest result for allocate_code: inconclusive!\n\n" )
line = fPETC()
if compare( line, "keep\n" ) != 0 then
    if platform() <= WIN32 then 
	system( "del wordwrap.log", 2 )
    else
        system( "rm wordwrap.log", 2 )
    end if
end if
ifdef UNITTEST then
	test_report()
end ifdef