-- xalloc.e:  Version of May 4th, 2008

include misc.e
include dll.e
include file.e
include machine.e
include get.e

-- Linux constants
constant PROT_EXEC = 4, PROT_READ = 1, PROT_WRITE = 2,
 PROT_NONE = 0

constant MAP_ANONYMOUS = #20, MAP_PRIVATE = #2
--,MAP_SHARED = #1, MAP_TYPE = #F, MAP_FIXED = #10,
--MAP_FILE = 0

-- Windows constants
constant MEM_COMMIT = #1000,
	MEM_RESERVE = #2000,
	--MEM_RESET = #8000,
	MEM_RELEASE = #8000

global constant 
        PAGE_EXECUTE = #10,
	PAGE_EXECUTE_READ = #20,
	PAGE_EXECUTE_READWRITE = #40,
	PAGE_EXECUTE_WRITECOPY = #80,
	PAGE_WRITECOPY = #08,
	PAGE_READWRITE = #04,
	PAGE_NOACCESS = #01,
	PAGE_READONLY = #02

atom kernel_dll, memDLL_id
atom getpagesize_rid, mmap_rid, VirtualAlloc_rid, VirtualLock_rid, VirtualProtect_rid, GetLastError_rid, GetSystemInfo_rid,
mprotect_rid, VirtualFree_rid, munmap_rid
atom openDLL_rid -- this must be a dynamic call or executing with the DOS version kills the program.



function xalloc_open_dll( sequence dllname ) 
	return call_func( openDLL_rid, { dllname } )
end function
function xalloc_opendll( sequence dllname)
	return xalloc_open_dll( dllname )
end function

if platform() != DOS32 then
	openDLL_rid = routine_id("open_dll")

	if platform() = WIN32 then
		memDLL_id = xalloc_open_dll( "kernel32.dll" )
		kernel_dll = memDLL_id
		VirtualAlloc_rid = define_c_func( memDLL_id, "VirtualAlloc", { C_POINTER, C_UINT, C_UINT, C_UINT }, C_POINTER )
		VirtualProtect_rid = define_c_func( memDLL_id, "VirtualProtect", { C_POINTER, C_UINT, C_INT, C_POINTER }, C_INT )
		VirtualLock_rid = define_c_func( memDLL_id, "VirtualLock", { C_POINTER, C_UINT }, C_UINT )
		GetLastError_rid = define_c_func( kernel_dll, "GetLastError", {}, C_UINT )
		GetSystemInfo_rid = define_c_proc( kernel_dll, "GetSystemInfo", { C_POINTER } )
		VirtualFree_rid = define_c_func( kernel_dll, "VirtualFree", { C_POINTER, C_UINT, C_INT }, C_UINT )
	elsif find( platform(), { FREEBSD, LINUX } ) then
		
		getpagesize_rid = define_c_func( -1, "getpagesize", { }, C_UINT )
		mmap_rid = define_c_func( -1, "mmap", { C_POINTER, C_UINT, C_INT, C_INT, C_INT, C_INT }, C_POINTER )
		mprotect_rid = define_c_func( -1, "mprotect", { C_POINTER, C_UINT, C_INT }, C_INT )
		munmap_rid = define_c_func( -1, "munmap", { C_POINTER, C_UINT }, C_INT )

	end if
else
	
end if

function VirtualAlloc( atom addr, atom size, atom flallocationtype, atom flprotect )
	atom r1
	r1 = c_func( VirtualAlloc_rid, {addr, size, flallocationtype, flprotect } )
	return r1
end function

function mmap( object start, integer length, integer protection, integer flags, integer fd, integer offset )
	atom pc
	if atom( start ) then
		return c_func( mmap_rid, { start, length, protection, flags, fd, offset } )
	else
		pc = mmap( 0, length, protection, flags, fd, offset )
		poke( pc, start )
		return pc
	end if
end function


function mem_const_set( sequence s, sequence hash )
	for i = 1 to length( hash ) do
		s[log(hash[i][1])/log(2)+1] = hash[i][2]
	end for
	return s
end function

function make_constants_table()
	sequence s
	s  =  repeat( PROT_NONE, 10 )
	s = mem_const_set( s, { 
		{ PAGE_EXECUTE_READWRITE, or_bits( PROT_READ, or_bits( PROT_EXEC, PROT_WRITE ) ) },
		{ PAGE_EXECUTE_READ, or_bits( PROT_READ, PROT_EXEC ) },
		{ PAGE_EXECUTE, PROT_EXEC },
		{ PAGE_EXECUTE_WRITECOPY, or_bits( PROT_READ, or_bits( PROT_EXEC, PROT_WRITE ) ) },
		{ PAGE_NOACCESS, PROT_NONE },
		{ PAGE_READONLY, PROT_READ },
		{ PAGE_READWRITE, or_bits( PROT_READ, PROT_WRITE ) },
		{ PAGE_WRITECOPY, or_bits( PROT_READ, PROT_WRITE ) }
	} )	
	return s
end function

global type valid_windows_memory_protection_constant( integer x )
	atom value
	value = log(x)/log(2)
	return integer( value ) and (value <= 8) and (value >= 0)
end type

constant mem_constants_table = make_constants_table()

function mem_win2linux( valid_windows_memory_protection_constant protection )
	return mem_constants_table[log(protection)/log(2)+1]
end function

type multiple_of_4( atom a )
	return not remainder( a, 4 )
end type

-- allocate_protect( data, integer protection )
-- allocates and copies data into memory and gives it protection using 
-- Microsoft's Memory Protection Constants.  The user may only pass in
-- one of these constants.  You may not pass in two that have been
-- bitwise ored together.
--
-- Microsoft's Memory Protection Constants
-- http://msdn.microsoft.com/en-us/library/aa366786(VS.85).aspx
--
-- The first parameter, data, is the machine code to be executed
--
-- The function returns the address to the required memory
-- or 0 if it fails.  This function is guaranteed to return memory on 
-- the 4 byte boundary.  It also guarantees that the memory returned with 
-- at least the protection given (but you may get more).
--
-- If you want to call allocate_protect( data, PAGE_READWRITE ), you can use 
-- allocate() instead.  It is more efficient.
--
-- If you want to call allocate_protect( data, PAGE_EXECUTE ), you can use 
-- allocate_code() instead.  It is simplier.
--
-- Use this one if you want something really unusual like you want to convert
-- a sequence to read-only values in memory.
-- 
--
-- You mustn't use free() on memory returned from this function.
global function allocate_protect( sequence data, valid_windows_memory_protection_constant protection )
	multiple_of_4 addr, oldprotptr
	integer size
	
	size = length(data)

	if platform() = DOS32 then

		addr = allocate( size )
		if addr = 0 then
			return 0
		end if
		poke( addr, data )

	elsif platform() = WIN32 then

		addr = c_func( VirtualAlloc_rid, { 0, size, or_bits( MEM_RESERVE, MEM_COMMIT ), PAGE_READWRITE } )
		if addr = 0 then
		    return 0
		end if
		oldprotptr = allocate(4)
		if oldprotptr = 0 then
		    return 0
		end if
		poke( addr, data )
		if c_func( VirtualProtect_rid, { addr, size, protection , oldprotptr } ) = 0 then
		    -- 0 indicates failure here
		    return 0
		end if
		free( oldprotptr )

	elsif platform() = LINUX then

		addr = c_func( mmap_rid, { 0, size, PROT_WRITE, or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if addr = -1 then
			return 0
		end if
		poke( addr, data )
		if c_func( mprotect_rid, { addr, size, mem_win2linux( protection ) } ) != 0 then
			-- non zero indicates failure here
			return 0
		end if

	else

		addr = allocate( size )
		poke( addr, data )

	end if

	return addr

	-- Implementation notes:
	--    The amount of memory actually allocated is the lowest multiple of the page_size (4kB on Windows XP) 
	--    The C manuals do not guarantee that the memory RETURNED from these pages aligned only what you pass into the
	--    underlyng C-functions must be page aligned.  I suspect that is the spirit of the work though.  So,
	--    I will make this thing crash if an address which is not a multiple of 4 bytes is returned.
end function

global function xallocate( integer size )
	return allocate_protect( repeat(0,size), PAGE_EXECUTE_READWRITE )
end function

-- allocate_code( data )
-- allocates and copies data into executible memory.
--
-- The first parameter, data, is the machine code to be executed
--
-- The function returns the address to the required memory
-- or 0 if it fails.  This function is guaranteed to return memory on the 4 byte
-- boundary.  It also guarantees that the memory returned will be executable.
--
-- The memory may not be readable or writable.  This function tries
-- to make the memory Execute Only.  If you want to also write
-- to this memory you should use allocate_protect instead.
--
-- You mustn't use free() on memory returned from this function.
global function allocate_code( sequence data )
	multiple_of_4 addr, oldprotptr
	integer size
	
	size = length(data)

	if platform() = DOS32 then

		addr = allocate( size )
		if addr = 0 then
			return 0
		end if
		poke( addr, data )

	elsif platform() = WIN32 then
		
		addr = VirtualAlloc( 0, size, or_bits( MEM_RESERVE, MEM_COMMIT ), PAGE_READWRITE )
		oldprotptr = allocate(4) 
		if addr = 0 then
		    return 0
		end if
		poke( addr, data )
		if c_func( VirtualProtect_rid, { addr, size, PAGE_EXECUTE , oldprotptr } ) = 0 then
			-- 0 indicates failure here
			return 0
		end if
		free( oldprotptr )
		
	elsif find( platform(), { FREEBSD, LINUX } ) then

		addr = c_func( mmap_rid, { 0, size, PROT_WRITE, or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if addr = -1 then
			return 0
		end if
		poke( addr, data )
		if c_func( mprotect_rid, { addr, size, PROT_EXEC } ) != 0 then
			-- non zero indicates failure here
			return 0
		end if

	else -- unknown platform.  Return normal memory.

		addr = allocate( size )
		poke( addr, data )

	end if

	return addr

	-- Implementation notes:
	--    The amount of memory actually allocated is the lowest multiple of the page_size (4kB on Windows XP) 
	--    The C manuals do not guarantee that the memory RETURNED from these pages aligned only what you 
	--    pass into the
	--    underlyng C-functions must be page aligned.  I suspect that is the spirit of the work though.  So,
	--    I will make this thing crash if an address which is not a multiple of 4 bytes is returned.
end function



-- new_addr = memory_reprotect( addr, size, protection )
--
-- Changes protection on a segment of memory using MS-WIndows style constants.
--
-- The memory should have been returned from allocate_protect or allocate_code.
--
-- The affected region is really:
--        floor( addr / page_size() ) * page_size() .. ceil( (addr+size) / page_size() ) * page_size
-- On an error 0 is returned.  If you ask this function to do what it can't it
-- will return 0.
-- See the memory protection constants at:
--   http://msdn2.microsoft.com/en-us/library/aa366786(VS.85).aspx
--   The idea is to support everything that both mmap and VirtualAlloc can
--   do and no more.
--
--
-- Implementation notes: The memory constants that are less than #100 are supported
-- and you may NOT or two together.
-- WARNING: UNTESTED FUNCTION
global function memory_reprotect( atom addr, integer size, valid_windows_memory_protection_constant protection )
	integer linux_protection
	atom new_addr
	if platform() = DOS32 then
		if addr = 0 then
			new_addr = allocate( size )
			poke( new_addr, peek( { addr, size } ) )
		else
		    	new_addr = addr
		end if
	elsif platform() = WIN32 then
		new_addr = VirtualAlloc( addr, size, or_bits( MEM_RESERVE, MEM_COMMIT ), protection )
		poke( new_addr, peek( { addr, size } ) )
	elsif find( platform(), { LINUX, FREEBSD } ) then
		
		linux_protection = mem_win2linux( protection )
		new_addr = c_func( mmap_rid, { 0, size, linux_protection, 
			or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if new_addr = -1 then
			return 0
		end if
	else
		return 0
	end if
	return new_addr
end function	
	
-- For freeing memory allocated by allocate_protect() or allocate_code()
-- Do not pass memory returned from allocate() here!
global procedure free_code( atom addr, integer size )
	integer free_succeeded
	if platform() = WIN32 then
		free_succeeded = c_func( VirtualFree_rid, { addr, size, MEM_RELEASE } )
	elsif find( platform(), { LINUX, FREEBSD } ) then
		free_succeeded = not c_func( munmap_rid, { addr, size } )
	elsif platform() = DOS32 then
		free( addr )
	end if
end procedure

constant call_back_size = 92 -- maximum value of C based Euphoria and the 
			     -- Euphoria based Euphoria.

-- Returns a 32-bit call-back address for a Euphoria routine
-- This routine replace call_back for those using Euphoria interpreters
-- that do not return executable memory addresses.
-- id can be of the form: 
--     routine_id          - for Linux or Windows stdcall calls, 
-- or 
--     {'+', routine_id}   - for Windows cdecl calls
global function xcall_back(object id)
    return allocate_code( peek( { call_back( id ), call_back_size } ) )
end function

-- Untested routine for locking memory into physical RAM.
-- A portable version for those who want to lock memory into RAM in Windows.
-- Memory gets locked into RAM if lock is True and the system supports it.
-- If it fails it returns GET_FAIL
-- If it succeeds it returns GET_SUCCESS
-- WARNING: UNTESTED FUNCTION
-- Implementation note: Will always fail on UNIX systems.  The UNIX part is
-- not yet ported.
function lock_memory( atom addr, integer size, integer lock )
	integer lockfailed
	-- for UNIX there is mlock()
	if platform() = WIN32 then

		if lock then
			lockfailed = c_func( VirtualLock_rid, { addr, size } )
		end if
		if lockfailed then
			return GET_FAIL
		else
			return GET_SUCCESS
		end if

	elsif find( platform(), { FREEBSD, LINUX } ) then

		--lockfailed = mlock(...)
		return GET_FAIL

	else

		return GET_FAIL

	end if

end function

global function xalloc_loaded()
	if platform() = WIN32 then
		return VirtualAlloc_rid = -1 or VirtualProtect_rid = -1 
			or GetLastError_rid = -1 or GetSystemInfo_rid = -1
	elsif platform() = LINUX then
		return mmap_rid = -1 or getpagesize_rid = -1
	end if
	return 0
end function	
