-- <!-- View this in a browser by renaming extension to html if necessary...
--> (c) Copyright 2008 Shawn Pringle</H1>                                  <!--  
--> <H1>xalloc.e:  Version of December 29th, 2008</H1><P>                  <!--
--> <H2>Description</H2> A memory protection API that is portable          <!--
-->              across several platforms.  Including                      <!--
-->              DOS, Windows, OSX, FreeBSD, and Linux.                    <!--
--> <BR> <!--
-->  The idea of this API is to give the programmer a method of telling the OS what he intends to do with various allocated parts of memory.  What does he intend to execute and what memory is only data.  This seems dubious at first, but if you read about hacking and security flaws you'll see that many exploits are done by tricking a program into executing a peice of memory that was intended to be only data.  Data Execute Protection stops that cold.  Now, when you are running Windows with DEP enabled, and a program attempts to execute data allocated by <font color="green"><nobr>allocate()</nobr></font> the program will be shutdown.  You don't have to do anything except enable it in the system and newer versions of XP and Vista already have it enabled by default.<P><!--
-->  The problem is, what if you want to execute dynamic code using <font color="green"><nobr>call()</nobr></font>, <font color="green">c_func()</font>, , <font color="green">c_proc()</font>, <font color="green">call_proc()</font>, <font color="green">call_func()</font>?  Is the memory address returned by <font color="green">call_back()</font> DEP safe?  It turns out, since C is already marked as executable and EUPHORIA routines are interpreted you only need to replace <font color="green">call_back</font> with <a href="#xcall_back"><font color="green"><nobr>xcall_back()</nobr></font></a> and you need to ensure that what you pass to <font color="green"><nobr>call()</nobr></font> is returned from <a href="#allocate_code"><font color="green">allocate_code()</font></a> and not from <nobr><font color="green">allocate()</font></nobr>.<P> <!--
-->  On Operating systems that do not support DEP this library simply wraps the standard generic memory routines: allocate() and free().  This library will work with either 4.0 or 3.1 versions of EUPHORIA and on any platform.<!--
include misc.e
include dll.e
include file.e
include machine.e
include get.e

-- This works on 4.0 too even though 4.0 defines OSX:
constant OSX = 4

-- Linux constants
constant PROT_EXEC = 4, PROT_READ = 1, PROT_WRITE = 2,
 PROT_NONE = 0

constant MAP_ANONYMOUS = #20, MAP_PRIVATE = #2
--,MAP_SHARED = #1, MAP_TYPE = #F, MAP_FIXED = #10,
--MAP_FILE = 0

-- Windows constants
constant MEM_COMMIT = #1000,
	MEM_RESERVE = #2000,
	--MEM_RESET = #8000,
	MEM_RELEASE = #8000
--> <H2>Constants</H2>Constants are to be sent to the API functions and are <!--
--> the same constants across all platforms.<br><PRE> <!--
global constant 
-->
	PAGE_EXECUTE = #10,        -- You may run the data in this page
	PAGE_EXECUTE_READ = #20,   -- You may run and read the data
	PAGE_EXECUTE_READWRITE = #40,-- You may run, read or write this page
	PAGE_EXECUTE_WRITECOPY = #80,-- You may run, read or write this page
	PAGE_WRITECOPY = #08,        
	PAGE_READWRITE = #04,
	PAGE_NOACCESS = #01,        -- 
	PAGE_READONLY = #02         -- You may only read code
-- </PRE><P><!--
atom kernel_dll, memDLL_id
atom getpagesize_rid, mmap_rid, VirtualAlloc_rid, 
	VirtualLock_rid, VirtualUnlock_rid, 
	VirtualProtect_rid, GetLastError_rid, 
	GetSystemInfo_rid,
	mprotect_rid, VirtualFree_rid, munmap_rid,
	mlock_rid, munlock_rid, VirtualQuery_rid
	
atom openDLL_rid -- this must be a dynamic call or executing with the DOS version kills the program.



function xalloc_open_dll( sequence dllname ) 
	return call_func( openDLL_rid, { dllname } )
end function
function xalloc_opendll( sequence dllname)
	return xalloc_open_dll( dllname )
end function

if platform() != DOS32 then
	openDLL_rid = routine_id("open_dll")

	if platform() = WIN32 then
		memDLL_id = xalloc_open_dll( "kernel32.dll" )
		kernel_dll = memDLL_id
		VirtualQuery_rid = define_c_func( memDLL_id, "VirtualQuery", { C_POINTER, C_POINTER, C_UINT }, C_UINT )
		VirtualAlloc_rid = define_c_func( memDLL_id, "VirtualAlloc", { C_POINTER, C_UINT, C_UINT, C_UINT }, C_POINTER )
		VirtualProtect_rid = define_c_func( memDLL_id, "VirtualProtect", { C_POINTER, C_UINT, C_INT, C_POINTER }, C_INT )
		VirtualLock_rid = define_c_func( memDLL_id, "VirtualLock", { C_POINTER, C_UINT }, C_UINT )
		VirtualUnlock_rid = define_c_func( memDLL_id, "VirtualUnlock", { C_POINTER, C_UINT }, C_UINT )
		GetLastError_rid = define_c_func( kernel_dll, "GetLastError", {}, C_UINT )
		GetSystemInfo_rid = define_c_proc( kernel_dll, "GetSystemInfo", { C_POINTER } )
		VirtualFree_rid = define_c_func( kernel_dll, "VirtualFree", { C_POINTER, C_UINT, C_INT }, C_UINT )
		
	elsif find( platform(), { FREEBSD, LINUX } ) then
		
		getpagesize_rid = define_c_func( -1, "getpagesize", { }, C_UINT )
		mmap_rid = define_c_func( -1, "mmap", { C_POINTER, C_UINT, C_INT, C_INT, C_INT, C_INT }, C_POINTER )
		mprotect_rid = define_c_func( -1, "mprotect", { C_POINTER, C_UINT, C_INT }, C_INT )
		munmap_rid = define_c_func( -1, "munmap", { C_POINTER, C_UINT }, C_INT )
		mlock_rid = define_c_func( -1, "mlock", { C_POINTER, C_UINT }, C_INT )
		munlock_rid = define_c_func( -1, "munlock", { C_POINTER, C_UINT }, C_INT )
		
	end if
else
	
end if

function VirtualAlloc( atom addr, atom size, atom flallocationtype, atom flprotect )
	atom r1
	r1 = c_func( VirtualAlloc_rid, {addr, size, flallocationtype, flprotect } )
	return r1
end function

function mmap( object start, integer length, integer protection, integer flags, integer fd, integer offset )
	atom pc
	if atom( start ) then
		return c_func( mmap_rid, { start, length, protection, flags, fd, offset } )
	else
		pc = mmap( 0, length, protection, flags, fd, offset )
		poke( pc, start )
		return pc
	end if
end function


function mem_const_set( sequence s, sequence hash )
	for i = 1 to length( hash ) do
		s[log(hash[i][1])/log(2)+1] = hash[i][2]
	end for
	return s
end function

function make_constants_table()
	sequence s
	s  =  repeat( PROT_NONE, 10 )
	s = mem_const_set( s, { 
		{ PAGE_EXECUTE_READWRITE, or_bits( PROT_READ, or_bits( PROT_EXEC, PROT_WRITE ) ) },
		{ PAGE_EXECUTE_READ, or_bits( PROT_READ, PROT_EXEC ) },
		{ PAGE_EXECUTE, PROT_EXEC },
		{ PAGE_EXECUTE_WRITECOPY, or_bits( PROT_READ, or_bits( PROT_EXEC, PROT_WRITE ) ) },
		{ PAGE_NOACCESS, PROT_NONE },
		{ PAGE_READONLY, PROT_READ },
		{ PAGE_READWRITE, or_bits( PROT_READ, PROT_WRITE ) },
		{ PAGE_WRITECOPY, or_bits( PROT_READ, PROT_WRITE ) }
	} )	
	return s
end function


global type valid_windows_memory_protection_constant( integer x )
	atom value
	value = log(x)/log(2)
	return integer( value ) and (value <= 8) and (value >= 0)
end type

global type page_aligned_address( atom a )
	return remainder( a, 4096 ) = 0
end type

constant mem_constants_table = make_constants_table()

function mem_win2linux( valid_windows_memory_protection_constant protection )
	return mem_constants_table[log(protection)/log(2)+1]
end function

type multiple_of_4( atom a )
	return not remainder( a, 4 )
end type

--> <H2>Routines</H2> <!--

-- atom allocate_protect( sequence data, integer protection )
--> <a name="allocate_protect"><h3><font color="red">atom</font> <font color="green">allocate_protect</font>( <font color="red">data</font>, <font color="red">integer</font> protection )</h3> <!--
--> allocates and copies data into memory and gives it protection using Microsoft's Memory Protection Constants.  The user may only pass in one of these constants.  If you only wish to execute a sequence as machine code use <a href="#allocate_code"><font color="green">allocate_code()</font></a>.  If you only want to read and write data into memory use <font color="green"><nobr>allocate()</nobr></font><P> <!--

--><a href="http://msdn.microsoft.com/en-us/library/aa366786(VS.85).aspx">Microsoft's Memory Protection Constants<br><!--
--> http://msdn.microsoft.com/en-us/library/aa366786(VS.85).aspx</a><p><!--

--><h4>Parameters</h4><!--
--> The first parameter, data, is the machine code to be put into memory. <!--
--
-- The function returns the address to the required memory
-- or 0 if it fails.  This function is guaranteed to return memory on 
-- the 4 byte boundary.  It also guarantees that the memory returned with 
-- at least the protection given (but you may get more).
--
-- If you want to call allocate_protect( data, PAGE_READWRITE ), you can use 
-- allocate() instead.  It is more efficient.
--
-- If you want to call allocate_protect( data, PAGE_EXECUTE ), you can use 
-- allocate_code() instead.  It is simplier.
--
-- Use this one if you want something really unusual like you want to convert
-- a sequence to read-only values in memory.
-- 
--
-- You mustn't use free() on memory returned from this function.
global function allocate_protect( sequence data, valid_windows_memory_protection_constant protection )
	atom addr, oldprotptr
	integer size
	
	size = length(data)

	if platform() = DOS32 or not xalloc_loaded() then

		addr = allocate( size )
		if addr = 0 then
			return 0
		end if
		poke( addr, data )

	elsif platform() = WIN32 then

		addr = c_func( VirtualAlloc_rid, { 0, size, or_bits( MEM_RESERVE, MEM_COMMIT ), PAGE_READWRITE } )
		if addr = 0 or not multiple_of_4( addr ) then
		    return 0
		end if
		oldprotptr = allocate(4)
		if oldprotptr = 0 then
		    return 0
		end if
		poke( addr, data )
		if c_func( VirtualProtect_rid, { addr, size, protection , oldprotptr } ) = 0 then
		    -- 0 indicates failure here
		    return 0
		end if
		free( oldprotptr )

	elsif find( platform(), {LINUX, FREEBSD, OSX} ) then

		addr = c_func( mmap_rid, { 0, size, PROT_WRITE, or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if addr = -1 then
			return 0
		end if
		poke( addr, data )
		if c_func( mprotect_rid, { addr, size, mem_win2linux( protection ) } ) != 0 then
			-- non zero indicates failure in mprotect 			
			return 0
		end if

	else

		addr = allocate( size )
		poke( addr, data )

	end if

	return addr

	-- Implementation notes:
	--    The amount of memory actually allocated on Windows is the lowest 
	--    multiple of the page_size (4kB on Windows XP)  
	--    The C manuals do not guarantee that the memory RETURNED from 
	--    underlying C-functions are page aligned only but they require 
	--    what you pass into them must be page aligned.  I suspect that 
	--    that is the spirit of the work though.  
end function

global function xallocate( integer size )
	atom addr
	
	if platform() = DOS32 or not xalloc_loaded() then

		addr = allocate( size )
		if addr = 0 then
			return 0
		end if

	elsif platform() = WIN32 then
		
		addr = VirtualAlloc( 0, size, or_bits( MEM_RESERVE, MEM_COMMIT ), PAGE_READWRITE )
		if addr = 0 then
		    return 0
		end if
		
	elsif find( platform(), { FREEBSD, LINUX, OSX } ) then

		addr = c_func( mmap_rid, { 0, size, or_bits(PROT_WRITE,PROT_EXEC), or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if addr = -1 then
			return 0
		end if

	else -- unknown platform.  Return normal memory.

		addr = allocate( size )

	end if

	return addr
	
end function

--> <a name="allocate_code"><H3><font color="red">atom</font> <font color="green">allocate_code</font>( <font color="red">sequence</font> a_sequence_of_machine_code_bytes )</H3>                    allocates and copies data into executible memory.<br>                           <h4>Parameters</h4><P>The parameter, a_sequence_of_machine_code_bytes, is the machine code to be put into memory.  To be later called with <nobr><font color="green">call()</font></nobr>            </P><!--

--><h4>Return Value</h4><P>The function returns the address in memory of the byte-code that can be safely executed whether DEP is enabled or not or 0 if it fails.  On the other hand, if you try to execute a code address returned by <nobr><font color="green">allocate()</font></nobr> with DEP enabled the program will receive a machine exception.<!--  

--> <h4>Comments</h4><P>Use this for the machine code you want to run in memory.  The copying is done for you and when the routine returns the memory may not be readable or writable but it is guaranteed to be executable.  If you want to also write to this memory <b>after the machine code has been copied</b> you should use <font color="green">allocate_protect()</font> instead and you should read about having memory executable and writable at the same time is a bad idea.  You mustn't use <nobr><font color="green">free()</font></nobr> on memory returned from this function.  </P>   <!--
global function allocate_code( sequence data )
	atom addr, oldprotptr
	integer size
	
	size = length(data)

	if platform() = DOS32 or not xalloc_loaded() then

		addr = allocate( size )
		if addr = 0 then
			return 0
		end if
		poke( addr, data )

	elsif platform() = WIN32 then
		
		addr = VirtualAlloc( 0, size, or_bits( MEM_RESERVE, MEM_COMMIT ), PAGE_READWRITE )
		oldprotptr = allocate(4) 
		if addr = 0 then
		    return 0
		end if
		poke( addr, data )
		if c_func( VirtualProtect_rid, { addr, size, PAGE_EXECUTE , oldprotptr } ) = 0 then
			-- 0 indicates failure here
			return 0
		end if
		free( oldprotptr )
		
	elsif find( platform(), { FREEBSD, LINUX, OSX } ) then

		addr = c_func( mmap_rid, { 0, size, PROT_WRITE, or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if addr = -1 then
			return 0
		end if
		poke( addr, data )
		if c_func( mprotect_rid, { addr, size, PROT_EXEC } ) != 0 then
			-- non zero indicates failure here
			return 0
		end if

	else -- unknown platform.  Return normal memory.

		addr = allocate( size )
		poke( addr, data )

	end if

	return addr

end function



-- new_addr = memory_reprotect( addr, size, protection )
--
-- Changes protection on a segment of memory using MS-Windows style constants.
--
-- The memory should have been returned from allocate_protect or allocate_code.
--
-- The affected region is really:
--        floor( addr / page_size() ) * page_size() .. ceil( (addr+size) / page_size() ) * page_size
-- On an error 0 is returned.  If you ask this function to do what it can't it
-- will return 0.
-- See the memory protection constants at:
--   http://msdn2.microsoft.com/en-us/library/aa366786(VS.85).aspx
--   The idea is to support everything that both mmap and VirtualAlloc can
--   do and no more.
--
--
-- Implementation notes: The memory constants that are less than #100 are supported
-- and you may NOT put two together with or_bits().
-- WARNING: UNTESTED FUNCTION
global function memory_reprotect( page_aligned_address addr, integer size, valid_windows_memory_protection_constant protection )
	integer linux_protection
	atom new_addr
	
	if not xalloc_loaded() then
		return addr
	elsif platform() = DOS32 then
		if addr = 0 then
			new_addr = allocate( size )
			poke( new_addr, peek( { addr, size } ) )
		else
		    	new_addr = addr
		end if
	elsif platform() = WIN32 then
		new_addr = VirtualAlloc( addr, size, or_bits( MEM_RESERVE, MEM_COMMIT ), protection )
		poke( new_addr, peek( { addr, size } ) )
	elsif find( platform(), { LINUX, FREEBSD, OSX } ) then
		linux_protection = mem_win2linux( protection )
		new_addr = c_func( mmap_rid, { 0, size, linux_protection, 
			or_bits( MAP_PRIVATE, MAP_ANONYMOUS ), 0, 0 } )
		if new_addr = -1 then
			return 0
		end if
	else
		return 0
	end if
	return new_addr
end function	


--> <h3><font color="green">free_code</font>( <font color="red">atom</font> addr, <font color="red">integer</font> size )</h3> <P>frees up allocated code memory</P><!--
--><h4>Parameters</h4><!--
-->addr must be an address returned by <a href="#allocate_code"><nobr><font color="green">allocate_code()</font></nobr></a> or <a href="#allocate_protect"><nobr><font color="green">allocate_protect()</font></nobr></a>.  Do <b>not</b> pass memory returned from <font color="green">allocate()</font> here!   The size is the size you specified when you called the afore mentioned functions.                           <h4>Comments</h4>                 <!--
-->Chances are you will not need to call this function because code allocations are typically global scope operations that you want to have available until your process exits. <!--
global procedure free_code( atom addr, integer size )
	integer free_succeeded
	if not xalloc_loaded() then
		free( addr )
	elsif platform() = WIN32 then
		free_succeeded = c_func( VirtualFree_rid, { addr, size, MEM_RELEASE } )
	elsif find( platform(), { LINUX, FREEBSD, OSX } ) then
		free_succeeded = not c_func( munmap_rid, { addr, size } )
	elsif platform() = DOS32 then
		free( addr )
	end if
end procedure

constant call_back_size = 92 -- maximum value of C based Euphoria and the 
			     -- Euphoria based Euphoria.

function memory_protection( atom addr )
	atom memory_basic_information_ptr
	atom protection
	memory_basic_information_ptr = allocate( 28 )
	if c_func( VirtualQuery_rid, { addr, memory_basic_information_ptr, call_back_size } ) < 12 then
		return -1
	end if
	protection = peek4u( memory_basic_information_ptr + 8 )
	free( memory_basic_information_ptr )
	return or_bits( protection, #FF )
end function
--> <a name="xcall_back"><h3><font color="red">atom</font> <font color="green">xcall_back</font>( <font color="red">object</font> routine_identifier )</h3> <!--			     
--><h4>Parameters</h4><P><!--
-->routine_identifier can be of the form:<UL> <!-- 
--><LI>     id         - for Linux or Windows stdcall calls,<!-- 
--><LI>    {'+', id}   - for Windows cdecl calls<!--
--></UL><!--
--><P>where id is a routine integer identifier returned from <font color="green">routine_id()</font>.<!--
--><h4>Return Value</h4> The function returns a call-back address of a EUPHORIA routine that can be safely executed whether DEP is enabled or not.  On the other hand, if you try to execute a code address returned by <font color="green">call_back()</font> with DEP enabled the program will receive a machine exception.<P><!--
--><h4>Comments</h4>                                                             <p>This routine is a drop-in replacement for <font color="green">call_back()</font>.  Use search and replace in your favorite code editor to upgrade.</p> The routine is implemented by copying call_back() data into executable memory all of the non-DEP related comments that apply to <font color="green">call_back()</font> also apply to <font color="green">xcall_back()</font>.<!--
global function xcall_back(object id)
	atom cb_addr
	cb_addr = call_back( id )
	if  
	-- First see if the functions were loaded from the dll.  Second,
	-- If the user has patched the interpreter and is using xcall_back then
	-- the memory will not be readable.  So, we have to check that, too
	if xalloc_loaded() and 
	or_bits( PROT_READ, mem_win2linux( memory_protection( cb_addr ) ) ) = PROT_READ then 
		return allocate_code( peek( { cb_addr, call_back_size } ) )
	else
		return cb_addr
	end if
end function

-- Untested routine for unlocking memory from physical RAM.
-- A portable version for those who want to lock memory into RAM.
-- Memory gets unlocked from RAM.
-- If it fails it returns GET_FAIL
-- If it succeeds it returns GET_SUCCESS
-- WARNING: UNTESTED FUNCTION
-- Implementation note: Will always fail on DOS32 systems.
function unlock_memory( atom addr, integer size )
	integer unlockfailed
	-- for UNIX there is mlock()
	if platform() = WIN32 then

		unlockfailed = c_func( VirtualUnlock_rid, { addr, size } )
		if unlockfailed then
			return GET_FAIL
		else
			return GET_SUCCESS
		end if

	elsif find( platform(), { FREEBSD, LINUX, OSX } ) then

		unlockfailed = c_func( munlock_rid, { addr, size } )
		if unlockfailed then 
			return GET_FAIL
		else
			return GET_SUCCESS
		end if

	else

		return GET_FAIL

	end if

end function


-- Untested routine for locking memory into physical RAM.
-- A portable version for those who want to lock memory into RAM in Windows.
-- Memory gets locked into RAM if the system supports it.
-- If it fails it returns GET_FAIL
-- If it succeeds it returns GET_SUCCESS
-- WARNING: UNTESTED FUNCTION
-- Implementation note: Will always fail on DOS32 systems.
function lock_memory( atom addr, integer size )
	integer lockfailed
	atom nonDOSaddr

	if platform() = WIN32 then

		nonDOSaddr = addr
		lockfailed = c_func( VirtualLock_rid, { nonDOSaddr, size } )
		if lockfailed then
			return GET_FAIL
		else
			return GET_SUCCESS
		end if

	elsif find( platform(), { FREEBSD, LINUX, OSX } ) then

		nonDOSaddr = addr
		lockfailed = c_func( mlock_rid, { nonDOSaddr, size } )
		if lockfailed then 
			return GET_FAIL
		else
			return GET_SUCCESS
		end if

	else

		return GET_FAIL

	end if

end function

global function xalloc_loaded()
	if platform() = WIN32 then
		return VirtualAlloc_rid != -1 and VirtualProtect_rid != -1 
			and GetLastError_rid != -1 and GetSystemInfo_rid != -1
	elsif find( platform(), { LINUX, FREEBSD, OSX } ) then
		return mmap_rid != -1 and getpagesize_rid != -1
	else
		return 1
	end if
	return 0
end function	
-->
